from typing import Any

from pydantic import BaseModel, Field, field_validator


class SpiderSettings(BaseModel):
    url: str | None = None


class CrawlRequest(BaseModel):
    spider_name: str
    spider_config: SpiderSettings | None = None

    crawl_args: dict[str, Any] = Field(
        default=dict(),
        description="Additional arguments for the spider",
    )
    start_requests: bool = Field(
        default=False,
        description="Whether to use spider's start_requests method",
    )

    @field_validator("spider_config", mode="before")
    @classmethod
    def validate_request_or_start_requests(cls, v, info):
        if not v and not info.data.get("start_requests"):
            raise ValueError(
                "Either 'request.url' or 'start_requests' must be provided"
            )
        return v


class CrawlResponse(BaseModel):
    items: list[dict[str, Any]] = []
    errors: list[dict[str, Any]] = []

    stats: dict[str, Any] = {}



from asyncio import wait_for

from crochet import setup as setup_crochet
from scrapy import signals
from scrapy.crawler import CrawlerRunner
from scrapy.signalmanager import dispatcher
from scrapy.utils.defer import deferred_to_future

from scrapyrta.app.core.config import app_settings
from scrapyrta.app.core.config.spider_settings import get_spider_settings
from scrapyrta.app.core.exceptions import (
    SpiderRunError,
)
from scrapyrta.app.core.log import configure_logging, logger
from scrapyrta.app.core.utils import get_spider_class
from scrapyrta.app.schemas.crawl import CrawlRequest, CrawlResponse

setup_crochet()
configure_logging()

runner_settings = get_spider_settings()


async def run_spider(crawl_request: CrawlRequest) -> CrawlResponse:
    try:
        spider_cls = get_spider_class(crawl_request.spider_name)

        spider_kwargs = {}
        if not crawl_request.start_requests:
            if not crawl_request.spider_config or not crawl_request.spider_config.url:
                raise ValueError("URL is required when start_requests is False")

            spider_kwargs = crawl_request.spider_config.model_dump(exclude_none=True)
            spider_kwargs["start_urls"] = [crawl_request.spider_config.url]

        spider_kwargs.update(crawl_request.crawl_args)  # type: ignore

        results, errors = [], []

        def handle_item(item, response, spider):
            if item:
                results.append(item)

        def handle_event(failure, **kwargs):
            if failure:
                errors.append({"failure": failure.getErrorMessage()})

        dispatcher.connect(handle_item, signal=signals.item_passed)
        dispatcher.connect(handle_event, signal=signals.item_error)
        dispatcher.connect(handle_event, signal=signals.spider_error)

        runner = CrawlerRunner(settings=runner_settings)
        deferred = runner.crawl(spider_cls, **spider_kwargs)

        crawler = next(iter(runner.crawlers))

        await wait_for(
            deferred_to_future(deferred),
            timeout=app_settings.TIMEOUT_LIMIT,
        )

        return CrawlResponse(
            items=results,
            errors=errors,
            stats=crawler.stats.get_stats(),
        )

    except TimeoutError as e:
        err_txt = "TimeoutError: The spider took too long to respond."
        logger.error(err_txt)

        raise SpiderRunError(crawl_request.spider_name, err_txt) from e

    except Exception as e:
        err_txt = f"{type(e).__name__}: {e}" if str(e) else f"{type(e).__name__}."
        logger.error(err_txt)

        raise SpiderRunError(crawl_request.spider_name) from e



import os
from collections.abc import Callable, Iterator
from typing import TypeVar

from environ import Env

_T = TypeVar("_T")


def _walk_to_root(path: str) -> Iterator[str]:
    """
    Yield directories starting from the given directory up to the root
    """

    if not os.path.exists(path):
        raise OSError("Starting path not found")

    if os.path.isfile(path):
        path = os.path.dirname(path)

    last_dir = None
    current_dir = os.path.abspath(path)
    while last_dir != current_dir:
        yield current_dir
        parent_dir = os.path.abspath(os.path.join(current_dir, os.path.pardir))
        last_dir, current_dir = current_dir, parent_dir


def find_dotenv(filename: str = ".env") -> str | None:
    path = os.getcwd()

    for dirname in _walk_to_root(path):
        check_path = os.path.join(dirname, filename)
        if os.path.isfile(check_path):
            return check_path


def get_env(name: str, default: _T, type_func: Callable) -> _T:
    return type_func(f"{PROJECT_NAME.upper()}_{name}", default=default)


env = Env()
env.read_env(env_file=find_dotenv())

PROJECT_NAME = "ScrapyRTA"

DEBUG = get_env("DEBUG", default=False, type_func=env.bool)
LOG_LEVEL = get_env("LOG_LEVEL", default="INFO", type_func=env.str)
ENABLE_OPEN_API = get_env("ENABLE_OPEN_API", default=True, type_func=env.bool)

PROJECT_SETTINGS = None

TIMEOUT_LIMIT = get_env("TIMEOUT_LIMIT", default=30, type_func=env.int)  # seconds

LOG_ROTATION = get_env(
    "LOG_ROTATION",
    default="10 MB",
    type_func=env.str,
)  # e.g. "10 MB" or "1 day"

LOG_RETENTION = get_env(
    "LOG_RETENTION",
    default="1 week",
    type_func=env.str,
)  # e.g. "1 week" or "10 days"



from fastapi import APIRouter

from scrapyrta.app.api.v1 import router as v1_router
from scrapyrta.app.core.config import app_settings

router = APIRouter(prefix="")

router.include_router(v1_router)

if app_settings.DEBUG:

    @router.get("/")
    async def root():
        return {
            "message": f"Welcome to {app_settings.PROJECT_NAME} API",
            "docs": f"{router.prefix}/docs",
        }

    @router.get("/health")
    async def health_check():
        return {"status": "Healthy"}



from contextlib import asynccontextmanager

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

from scrapyrta.app.api import router as api_router
from scrapyrta.app.core.config import app_settings


@asynccontextmanager
async def lifespan(app: FastAPI):
    print(f"Starting {app_settings.PROJECT_NAME} API")
    yield
    print(f"Shutting down {app_settings.PROJECT_NAME} API")


def create_app() -> FastAPI:
    app = FastAPI(
        title=app_settings.PROJECT_NAME,
        description="Asynchronous HTTP API for running Scrapy spiders",
        version="0.1.0",
        docs_url="/docs" if app_settings.ENABLE_OPEN_API else None,
        lifespan=lifespan,
    )

    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_methods=["*"],
        allow_headers=["X-Requested-With"],
        allow_credentials=False,
    )

    app.include_router(api_router)

    return app



from scrapy.settings import Settings

from scrapyrta.app.core.config import app_settings
from scrapyrta.app.core.utils import get_project_settings


def get_spider_settings(custom_settings: dict | None = None) -> Settings:
    if custom_settings is None:
        custom_settings = {}

    settings = get_project_settings()

    default_settings = {
        "DEBUG": app_settings.DEBUG,
        "LOG_ENABLED": False,
        "TELNETCONSOLE_ENABLED": False,
        "EXTENSIONS": {
            "scrapy.webservice.WebService": None,
            "scrapy.extensions.logstats.LogStats": None,
            "scrapy.extensions.throttle.AutoThrottle": None,
        },
    }

    settings.update(default_settings, priority="project")
    settings.update(custom_settings, priority="spider")

    return settings



import threading
from typing import ClassVar


class Singleton(type):
    _instances: ClassVar[dict] = {}
    _lock = threading.Lock()  # to synchronize threads (if we had any)

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls not in cls._instances:
                cls._instances[cls] = super().__call__(*args, **kwargs)

        return cls._instances[cls]



from scrapy.settings import Settings
from scrapy.spiderloader import SpiderLoader
from scrapy.spiders import Spider

from scrapyrta.app.core.config import app_settings
from scrapyrta.app.core.exceptions import SpiderNotFoundError


def get_project_settings(
    module: str | None = None,
    custom_settings: dict | None = None,
) -> Settings:
    crawler_settings = Settings()

    if module is None:
        module = app_settings.PROJECT_SETTINGS

    if custom_settings is None:
        custom_settings = {}

    crawler_settings.setmodule(module, priority="project")

    if custom_settings:
        crawler_settings.setdict(custom_settings, priority="cmdline")

    return crawler_settings


def get_spider_class(spider_name: str) -> type[Spider]:
    try:
        spider_loader = SpiderLoader.from_settings(settings=get_project_settings())
        spider_class = spider_loader.load(spider_name)

        return spider_class

    except KeyError as e:
        raise SpiderNotFoundError(spider_name) from e



from fastapi import APIRouter, HTTPException

from scrapyrta.app.core.exceptions import SpiderNotFoundError, SpiderRunError
from scrapyrta.app.schemas.crawl import CrawlRequest, CrawlResponse
from scrapyrta.app.services import run_spider

router = APIRouter(prefix="")


@router.post("/crawl.json", response_model=CrawlResponse)
async def start_crawl(request: CrawlRequest) -> CrawlResponse:
    """
    Start a spider crawl with the specified parameters.

    The spider can be started in two ways:
    1. By providing a URL in the request
    2. By enabling start_requests
    """

    exception_handlers = {
        SpiderNotFoundError: 404,
        SpiderRunError: 500,
    }

    try:
        return await run_spider(crawl_request=request)

    except tuple(exception_handlers.keys()) as e:
        status_code = exception_handlers[type(e)]
        raise HTTPException(status_code=status_code, detail=str(e)) from e

    except Exception as e:
        raise HTTPException(
            status_code=500,
            detail=f"An unexpected error occurred: {e!s}",
        ) from e



from fastapi import HTTPException


class ScrapyRTAException(Exception):
    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)


class SpiderNotFoundError(ScrapyRTAException):
    def __init__(self, spider_name: str):
        self.spider_name = spider_name
        message = f"Spider '{spider_name}' not found"
        super().__init__(message)


class SpiderRunError(ScrapyRTAException):
    def __init__(self, spider_name: str, message: str | None = None):
        self.spider_name = spider_name

        if message is None:
            message = f"Error running spider '{spider_name}': check logs for details."
        else:
            message = f"Error running spider '{spider_name}': {message}"

        super().__init__(message)


def spider_exception_handler(exception: ScrapyRTAException) -> HTTPException:
    status_code_map: dict[type[ScrapyRTAException], int] = {
        SpiderNotFoundError: 404,
        SpiderRunError: 500,
    }

    status_code = status_code_map.get(type(exception), 500)
    detail = (
        exception.message
        if isinstance(exception, ScrapyRTAException)
        else str(exception)
    )

    return HTTPException(status_code=status_code, detail=detail)



import argparse
import os
import sys
from configparser import ConfigParser, NoOptionError, NoSectionError

import uvicorn
from scrapy.utils.conf import closest_scrapy_cfg

from scrapyrta.app.app import create_app
from scrapyrta.app.core.config import Settings, app_settings


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Async HTTP API server for Scrapy project.",
    )

    parser.add_argument(
        "-p",
        "--port",
        dest="port",
        type=int,
        default=9080,
        help="Port number to listen on",
    )
    parser.add_argument(
        "-i",
        "--ip",
        dest="ip",
        default="localhost",
        help="IP address the server will listen on",
    )
    parser.add_argument(
        "--project",
        dest="project",
        default="default",
        help="Project name from scrapy.cfg",
    )
    parser.add_argument(
        "-S",
        "--settings",
        dest="settings",
        metavar="project.settings",
        help="Custom project settings module path",
    )

    return parser.parse_args()


def find_scrapy_project(project) -> str:
    project_config_path = closest_scrapy_cfg()
    if not project_config_path:
        raise RuntimeError("Cannot find scrapy.cfg file")

    project_config = ConfigParser()
    project_config.read(project_config_path)

    try:
        project_settings = project_config.get("settings", project)
    except (NoSectionError, NoOptionError) as e:
        raise RuntimeError(str(e)) from e

    if not project_settings:
        raise RuntimeError("Cannot find scrapy project settings")

    project_location = os.path.dirname(project_config_path)
    sys.path.append(project_location)

    return project_settings


def run_application(arguments: argparse.Namespace, app_settings: Settings) -> None:
    app = create_app()
    app_settings.freeze()

    uvicorn.run(
        app,
        host=arguments.ip,
        port=arguments.port,
    )


def execute():
    sys.path.insert(0, os.getcwd())
    arguments = parse_arguments()

    if arguments.settings:
        app_settings.setmodule(arguments.settings)

    app_settings.set(
        "PROJECT_SETTINGS",
        find_scrapy_project(arguments.project),
    )

    if app_settings.DEBUG:
        print("Running in DEBUG mode...")

    run_application(arguments, app_settings)


def run():
    try:
        execute()

    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)



from copy import deepcopy
from importlib import import_module

from . import default_settings


class Settings:
    def __init__(self):
        self.setmodule(default_settings)

    def setmodule(self, module):
        if isinstance(module, str):
            module = import_module(module)

        for setting in dir(module):
            if setting.startswith("_") or not setting.isupper():
                continue

            self.set(setting, getattr(module, setting))

    def __setattr__(self, key, value):
        if self.frozen:
            raise TypeError("Trying to modify a frozen Settings object")

        return super().__setattr__(key, value)

    def __getattr__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError as e:
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute '{name}'"
            ) from e

    def set(self, name, value):
        if not name.startswith("_") and name.isupper():
            # Deepcopy objects here, or we will have issues with mutability,
            # like changing mutable object stored in settings leads to
            # change of object in default_settings module.
            setattr(self, name, deepcopy(value))

    def freeze(self):
        self._frozen = True

    def __str__(self) -> str:
        attrs = {
            key: getattr(self, key)
            for key in dir(self)
            if key.isupper() and not key.startswith("_")
        }

        formatted = "\n".join(f"{key}: {value}" for key, value in attrs.items())
        return f"{self.PROJECT_NAME} Settings:\n\n{formatted}\n"

    @property
    def frozen(self):
        return bool(getattr(self, "_frozen", False))


app_settings = Settings()



import logging
import sys
from pathlib import Path

from loguru import logger as loguru_logger

from scrapyrta.app.core.config import app_settings
from scrapyrta.app.utils.metaclasses import Singleton

BASE_DIR = Path(__file__).resolve(strict=True).parent.parent.parent


class LoggerConfig(metaclass=Singleton):
    def __init__(self, name: str = "scrapyrta"):
        self.name = name
        self.logger = self._configure_logger()

    def _configure_logger(self):
        # Remove default Loguru handler
        loguru_logger.remove()

        loguru_logger.add(
            sys.stdout,
            level=app_settings.LOG_LEVEL,
            serialize=True,
        )
        loguru_logger.add(
            BASE_DIR / "logs" / f"{self.name}.log",
            level=app_settings.LOG_LEVEL,
            rotation=app_settings.LOG_ROTATION,  # e.g. "10 MB" or "1 day"
            retention=app_settings.LOG_RETENTION,  # e.g. "1 week" or "10 days"
            serialize=True,
        )

        return loguru_logger.bind(name=self.name)

    def get_logger(self):
        return self.logger


def configure_logging():
    class LoguruLogger(logging.Handler):
        def emit(self, record):
            loguru_logger.opt(depth=1).log(record.levelname, record.getMessage())

    logging.getLogger("scrapy").addHandler(LoguruLogger())
    logging.getLogger("scrapy").setLevel(app_settings.LOG_LEVEL)


logger = LoggerConfig().get_logger()
